'use strict';

var require$$2 = require('combine-arrays');
var require$$3 = require('path-to-regexp-with-reversible-keys');
var require$$9 = require('then-denodeify');
var require$$10 = require('eventemitter3');
var require$$11 = require('hash-brown-router');
var require$$13 = require('page-path-builder');
var require$$14 = require('iso-next-tick');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var require$$2__default = /*#__PURE__*/_interopDefaultLegacy(require$$2);
var require$$3__default = /*#__PURE__*/_interopDefaultLegacy(require$$3);
var require$$9__default = /*#__PURE__*/_interopDefaultLegacy(require$$9);
var require$$10__default = /*#__PURE__*/_interopDefaultLegacy(require$$10);
var require$$11__default = /*#__PURE__*/_interopDefaultLegacy(require$$11);
var require$$13__default = /*#__PURE__*/_interopDefaultLegacy(require$$13);
var require$$14__default = /*#__PURE__*/_interopDefaultLegacy(require$$14);

var stateStringParser$2 = stateString => stateString.split(`.`).reduce((stateNames, latestNameChunk) => {
	stateNames.push(
		stateNames.length
			? stateNames[stateNames.length - 1] + `.` + latestNameChunk
			: latestNameChunk,
	);

	return stateNames
}, []);

const stateStringParser$1 = stateStringParser$2;

var stateState = function StateState() {
	const states = {};

	function getHierarchy(name) {
		const names = stateStringParser$1(name);

		return names.map(name => {
			if (!states[name]) {
				throw new Error(`State ${ name } not found`)
			}
			return states[name]
		})
	}

	function getParent(name) {
		const parentName = getParentName(name);

		return parentName && states[parentName]
	}

	function getParentName(name) {
		const names = stateStringParser$1(name);

		if (names.length > 1) {
			const secondToLast = names.length - 2;

			return names[secondToLast]
		} else {
			return null
		}
	}

	function guaranteeAllStatesExist(newStateName) {
		const stateNames = stateStringParser$1(newStateName);
		const statesThatDontExist = stateNames.filter(name => !states[name]);

		if (statesThatDontExist.length > 0) {
			throw new Error(`State ${ statesThatDontExist[statesThatDontExist.length - 1] } does not exist`)
		}
	}

	function buildFullStateRoute(stateName) {
		return getHierarchy(stateName).map(state => `/${ state.route || `` }`)
			.join(``)
			.replace(/\/{2,}/g, `/`)
	}

	function applyDefaultChildStates(stateName) {
		const state = states[stateName];

		const defaultChildStateName = state && (
			typeof state.defaultChild === `function`
				? state.defaultChild()
				: state.defaultChild
		);

		if (!defaultChildStateName) {
			return stateName
		}

		const fullStateName = `${ stateName }.${ defaultChildStateName }`;

		return applyDefaultChildStates(fullStateName)
	}


	return {
		add(name, state) {
			states[name] = state;
		},
		get(name) {
			return name && states[name]
		},
		getHierarchy,
		getParent,
		getParentName,
		guaranteeAllStatesExist,
		buildFullStateRoute,
		applyDefaultChildStates,
	}
};

var extend$2 = (...args) => Object.assign({}, ...args);

const stateStringParser = stateStringParser$2;
const extend$1 = extend$2;

const combine$1 = require$$2__default['default'];
const pathToRegexp = require$$3__default['default'];

var stateComparison_1 = function StateComparison(stateState) {
	const getPathParameters = pathParameters();

	const parametersChanged = args => parametersThatMatterWereChanged(extend$1(args, { stateState, getPathParameters }));

	return args => stateComparison(extend$1(args, { parametersChanged }))
};

function pathParameters() {
	const parameters = {};

	return path => {
		if (!path) {
			return []
		}

		if (!parameters[path]) {
			parameters[path] = pathToRegexp(path).keys.map(key => key.name);
		}

		return parameters[path]
	}
}

function parametersThatMatterWereChanged({ stateState, getPathParameters, stateName, fromParameters, toParameters }) {
	const state = stateState.get(stateName);
	const querystringParameters = state.querystringParameters || [];
	const parameters = getPathParameters(state.route).concat(querystringParameters);

	return Array.isArray(parameters) && parameters.some(
		key => fromParameters[key] !== toParameters[key],
	)
}

function stateComparison({ parametersChanged, original, destination }) {
	const states = combine$1({
		start: stateStringParser(original.name),
		end: stateStringParser(destination.name),
	});

	return states.map(({ start, end }) => ({
		nameBefore: start,
		nameAfter: end,
		stateNameChanged: start !== end,
		stateParametersChanged: start === end && parametersChanged({
			stateName: start,
			fromParameters: original.parameters,
			toParameters: destination.parameters,
		}),
	}))
}

var currentState = function CurrentState() {
	let current = {
		name: ``,
		parameters: {},
	};

	return {
		get() {
			return current
		},
		set(name, parameters) {
			current = {
				name,
				parameters,
			};
		},
	}
};

var stateChangeLogic$1 = function stateChangeLogic(stateComparisonResults) {
	let hitDestroyedState = false;

	const output = {
		destroy: [],
		create: [],
	};

	stateComparisonResults.forEach(state => {
		hitDestroyedState = hitDestroyedState || state.stateNameChanged || state.stateParametersChanged;

		if (state.nameBefore && hitDestroyedState) {
			output.destroy.push(state.nameBefore);
		}

		if (state.nameAfter && hitDestroyedState) {
			output.create.push(state.nameAfter);
		}
	});

	return output
};

var stateTransitionManager = emitter => {
	let currentTransitionAttempt = null;
	let nextTransition = null;

	function doneTransitioning() {
		currentTransitionAttempt = null;
		if (nextTransition) {
			beginNextTransitionAttempt();
		}
	}

	const isTransitioning = () => !!currentTransitionAttempt;

	function beginNextTransitionAttempt() {
		currentTransitionAttempt = nextTransition;
		nextTransition = null;
		currentTransitionAttempt.beginStateChange();
	}

	function cancelCurrentTransition() {
		currentTransitionAttempt.transition.cancelled = true;
		const err = new Error(`State transition cancelled by the state transition manager`);
		err.wasCancelledBySomeoneElse = true;
		emitter.emit(`stateChangeCancelled`, err);
	}

	emitter.on(`stateChangeAttempt`, beginStateChange => {
		nextTransition = createStateTransitionAttempt(beginStateChange);

		if (isTransitioning() && currentTransitionAttempt.transition.cancellable) {
			cancelCurrentTransition();
		} else if (!isTransitioning()) {
			beginNextTransitionAttempt();
		}
	});

	emitter.on(`stateChangeError`, doneTransitioning);
	emitter.on(`stateChangeCancelled`, doneTransitioning);
	emitter.on(`stateChangeEnd`, doneTransitioning);

	function createStateTransitionAttempt(beginStateChange) {
		const transition = {
			cancelled: false,
			cancellable: true,
		};
		return {
			transition,
			beginStateChange: (...args) => beginStateChange(transition, ...args),
		}
	}
};

var defaultRouterOptions$1 = { reverse: false };

// Pulled from https://github.com/joliss/promise-map-series and prettied up a bit

var promiseMapSeries = function sequence(array, iterator) {
	let currentPromise = Promise.resolve();
	return Promise.all(
		array.map((value, i) => currentPromise = currentPromise.then(() => iterator(value, i, array))),
	)
};

const StateState = stateState;
const StateComparison = stateComparison_1;
const CurrentState = currentState;
const stateChangeLogic = stateChangeLogic$1;
const parse = stateStringParser$2;
const StateTransitionManager = stateTransitionManager;
const defaultRouterOptions = defaultRouterOptions$1;

const series = promiseMapSeries;
const extend = extend$2;

const denodeify = require$$9__default['default'];
const EventEmitter = require$$10__default['default'];
const newHashBrownRouter = require$$11__default['default'];
const combine = require$$2__default['default'];
const buildPath = require$$13__default['default'];
const nextTick = require$$14__default['default'];

const getProperty = name => obj => obj[name];
const reverse = ary => ary.slice().reverse();
const isFunction = property => obj => typeof obj[property] === `function`;
const isThenable = object => object && (typeof object === `object` || typeof object === `function`) && typeof object.then === `function`;
const promiseMe = (fn, ...args) => new Promise(resolve => resolve(fn(...args)));

const expectedPropertiesOfAddState = [ `name`, `route`, `defaultChild`, `data`, `template`, `resolve`, `activate`, `querystringParameters`, `defaultQuerystringParameters`, `defaultParameters`, `canLeaveState` ];

var abstractStateRouter = function StateProvider(makeRenderer, rootElement, stateRouterOptions = {}) {
	const prototypalStateHolder = StateState();
	const lastCompletelyLoadedState = CurrentState();
	const lastStateStartedActivating = CurrentState();
	const stateProviderEmitter = new EventEmitter();
	const compareStartAndEndStates = StateComparison(prototypalStateHolder);

	const stateNameToArrayofStates = stateName => parse(stateName).map(prototypalStateHolder.get);

	StateTransitionManager(stateProviderEmitter);
	const { throwOnError, pathPrefix } = extend({
		throwOnError: true,
		pathPrefix: `#`,
	}, stateRouterOptions);

	const router = stateRouterOptions.router || newHashBrownRouter(defaultRouterOptions);

	router.on(`not found`, (route, parameters) => {
		stateProviderEmitter.emit(`routeNotFound`, route, parameters);
	});

	let destroyDom = null;
	let getDomChild = null;
	let renderDom = null;

	let activeStateResolveContent = {};
	const activeDomApis = {};
	const activeEmitters = {};

	function handleError(event, err) {
		nextTick(() => {
			stateProviderEmitter.emit(event, err);
			console.error(`${ event } - ${ err.message }`);
			if (throwOnError) {
				throw err
			}
		});
	}

	function destroyStateName(stateName) {
		const state = prototypalStateHolder.get(stateName);
		stateProviderEmitter.emit(`beforeDestroyState`, {
			state,
			domApi: activeDomApis[stateName],
		});

		activeEmitters[stateName].emit(`destroy`);
		activeEmitters[stateName].removeAllListeners();
		delete activeEmitters[stateName];
		delete activeStateResolveContent[stateName];

		return destroyDom(activeDomApis[stateName]).then(() => {
			delete activeDomApis[stateName];
			stateProviderEmitter.emit(`afterDestroyState`, {
				state,
			});
		})
	}

	function getChildElementForStateName(stateName) {
		return new Promise(resolve => {
			const parent = prototypalStateHolder.getParent(stateName);
			if (parent) {
				resolve(getDomChild(activeDomApis[parent.name]).then(childDomApi => {
					if (!childDomApi) {
						return Promise.reject(new Error(`getDomChild returned a falsey element, did you forget to add a place for a child state to go?`))
					}
					return childDomApi
				}));
			} else {
				resolve(rootElement);
			}
		})
	}

	function renderStateName(parameters, stateName) {
		return getChildElementForStateName(stateName).then(element => {
			const state = prototypalStateHolder.get(stateName);
			const content = getContentObject(activeStateResolveContent, stateName);

			stateProviderEmitter.emit(`beforeCreateState`, {
				state,
				content,
				parameters,
			});

			return renderDom({
				template: state.template,
				element,
				content,
				parameters,
			}).then(domApi => {
				activeDomApis[stateName] = domApi;
				stateProviderEmitter.emit(`afterCreateState`, {
					state,
					domApi,
					content,
					parameters,
				});
				return domApi
			})
		})
	}

	function renderAll(stateNames, parameters) {
		return series(stateNames, stateName => renderStateName(parameters, stateName))
	}

	function statesAreEquivalent(stateA, stateB) {
		const { create, destroy } = stateChangeLogic(
			compareStartAndEndStates({
				original: stateA,
				destination: stateB,
			}),
		);

		return create.length === 0 && destroy.length === 0
	}

	function allowStateChangeOrRevert(newStateName, newParameters) {
		const lastState = lastCompletelyLoadedState.get();
		if (lastState.name && statesAreEquivalent(lastState, lastStateStartedActivating.get())) {
			// Check canLeaveState for all states that will be destroyed or changed
			const { create, destroy } = stateChangeLogic(
				compareStartAndEndStates({
					original: lastState,
					destination: {
						name: newStateName,
						parameters: newParameters,
					},
				}),
			);
			const statesNamesToCheck = Array.from(new Set([ ...create, ...destroy ]).values());
			const canLeaveState = statesNamesToCheck.every(stateName => {
				const state = prototypalStateHolder.get(stateName);
				if (state?.canLeaveState && typeof state.canLeaveState === 'function') {
					const stateChangeAllowed = state.canLeaveState(activeDomApis[stateName]);
					if (!stateChangeAllowed) {
						stateProviderEmitter.emit('stateChangePrevented', stateName);
					}
					return stateChangeAllowed
				}
				return true
			});

			if (!canLeaveState) {
				stateProviderEmitter.go(lastState.name, lastState.parameters, { replace: true });
			}
			return canLeaveState
		}
		return true
	}

	function onRouteChange(state, parameters) {
		try {
			const finalDestinationStateName = prototypalStateHolder.applyDefaultChildStates(state.name);

			if (finalDestinationStateName === state.name && allowStateChangeOrRevert(state.name, parameters)) {
				emitEventAndAttemptStateChange(finalDestinationStateName, parameters);
			} else if (finalDestinationStateName !== state.name) {
				// There are default child states that need to be applied

				const theRouteWeNeedToEndUpAt = makePath(finalDestinationStateName, parameters);
				const currentRoute = router.location.get();

				if (theRouteWeNeedToEndUpAt !== currentRoute) {
					// change the url to match the full default child state route
					stateProviderEmitter.go(finalDestinationStateName, parameters, { replace: true });
				} else if (allowStateChangeOrRevert(finalDestinationStateName, parameters)) {
					// the child state has the same route as the current one, just start navigating there
					emitEventAndAttemptStateChange(finalDestinationStateName, parameters);
				}
			}
		} catch (err) {
			handleError(`stateError`, err);
		}
	}

	function addState(state) {
		if (typeof state === `undefined`) {
			throw new Error(`Expected 'state' to be passed in.`)
		} else if (typeof state.name === `undefined`) {
			throw new Error(`Expected the 'name' option to be passed in.`)
		} else if (typeof state.template === `undefined`) {
			throw new Error(`Expected the 'template' option to be passed in.`)
		}
		Object.keys(state).filter(key => expectedPropertiesOfAddState.indexOf(key) === -1).forEach(key => {
			console.warn(`Unexpected property passed to addState:`, key);
		});

		prototypalStateHolder.add(state.name, state);

		const route = prototypalStateHolder.buildFullStateRoute(state.name);

		router.add(route, parameters => onRouteChange(state, parameters));
	}

	function computeDefaultParams(defaultParams) {
		const computedDefaultParams = {};

		Object.keys(defaultParams).forEach(key => {
			computedDefaultParams[key] = typeof defaultParams[key] === `function` ? defaultParams[key]() : defaultParams[key];
		});

		return computedDefaultParams
	}

	function getStatesToResolve(stateChanges) {
		return stateChanges.create.map(prototypalStateHolder.get)
	}

	function emitEventAndAttemptStateChange(newStateName, parameters) {
		stateProviderEmitter.emit(`stateChangeAttempt`, function stateGo(transition) {
			attemptStateChange(newStateName, parameters, transition);
		});
	}

	function attemptStateChange(newStateName, parameters, transition) {
		function ifNotCancelled(fn) {
			return (...args) => {
				if (transition.cancelled) {
					const err = new Error(`The transition to ${ newStateName } was cancelled`);
					err.wasCancelledBySomeoneElse = true;
					throw err
				} else {
					return fn(...args)
				}
			}
		}

		return promiseMe(prototypalStateHolder.guaranteeAllStatesExist, newStateName)
			.then(function applyDefaultParameters() {
				const state = prototypalStateHolder.get(newStateName);
				const defaultParams = state.defaultParameters || state.defaultQuerystringParameters || {};
				const needToApplyDefaults = Object.keys(defaultParams).some(function missingParameterValue(param) {
					return typeof parameters[param] === `undefined`
				});

				if (needToApplyDefaults) {
					throw redirector(newStateName, extend(computeDefaultParams(defaultParams), parameters))
				}
				return state
			}).then(ifNotCancelled(state => {
				stateProviderEmitter.emit(`stateChangeStart`, state, parameters, stateNameToArrayofStates(state.name));
				lastStateStartedActivating.set(state.name, parameters);
			})).then(function getStateChanges() {
				const stateComparisonResults = compareStartAndEndStates({
					original: lastCompletelyLoadedState.get(),
					destination: {
						name: newStateName,
						parameters,
					},
				});
				return stateChangeLogic(stateComparisonResults) // { destroy, change, create }
			}).then(ifNotCancelled(function resolveDestroyAndActivateStates(stateChanges) {
				return resolveStates(getStatesToResolve(stateChanges), extend(parameters)).catch(function onResolveError(e) {
					e.stateChangeError = true;
					throw e
				}).then(ifNotCancelled(function destroyAndActivate(stateResolveResultsObject) {
					transition.cancellable = false;

					const activateAll = () => activateStates(
						stateChanges.create,
					);

					return series(reverse(stateChanges.destroy), destroyStateName).then(
						() => {
							activeStateResolveContent = extend(activeStateResolveContent, stateResolveResultsObject);
							return renderAll(stateChanges.create, extend(parameters)).then(activateAll)
						},
					)
				}))

				function activateStates(stateNames) {
					return stateNames.map(prototypalStateHolder.get).forEach(state => {
						const emitter = new EventEmitter();
						const context = Object.create(emitter);
						context.domApi = activeDomApis[state.name];
						context.data = state.data;
						context.parameters = parameters;
						context.content = getContentObject(activeStateResolveContent, state.name);
						activeEmitters[state.name] = emitter;

						try {
							state.activate && state.activate(context);
						} catch (e) {
							nextTick(() => {
								throw e
							});
						}
					})
				}
			})).then(function stateChangeComplete() {
				lastCompletelyLoadedState.set(newStateName, parameters);
				try {
					stateProviderEmitter.emit(`stateChangeEnd`, prototypalStateHolder.get(newStateName), parameters, stateNameToArrayofStates(newStateName));
				} catch (e) {
					handleError(`stateError`, e);
				}
			}).catch(ifNotCancelled(function handleStateChangeError(err) {
				if (err && err.redirectTo) {
					stateProviderEmitter.emit(`stateChangeCancelled`, err);
					return stateProviderEmitter.go(err.redirectTo.name, err.redirectTo.params, { replace: true })
				} else if (err) {
					handleError(`stateChangeError`, err);
				}
			})).catch(function handleCancellation(err) {
				if (err && err.wasCancelledBySomeoneElse) ; else {
					throw new Error(`This probably shouldn't happen, maybe file an issue or something ${ err }`)
				}
			})
	}

	function makePath(stateName, parameters, options) {
		function getGuaranteedPreviousState() {
			if (!lastStateStartedActivating.get().name) {
				throw new Error(`makePath required a previous state to exist, and none was found`)
			}
			return lastStateStartedActivating.get()
		}
		if (options && options.inherit) {
			parameters = extend(getGuaranteedPreviousState().parameters, parameters);
		}

		const destinationStateName = stateName === null ? getGuaranteedPreviousState().name : stateName;

		const destinationState = prototypalStateHolder.get(destinationStateName) || {};
		const defaultParams = destinationState.defaultParameters || destinationState.defaultQuerystringParameters || {};

		parameters = extend(computeDefaultParams(defaultParams), parameters);

		prototypalStateHolder.guaranteeAllStatesExist(destinationStateName);
		const route = prototypalStateHolder.buildFullStateRoute(destinationStateName);
		return buildPath(route, parameters || {})
	}

	const defaultOptions = {
		replace: false,
	};

	stateProviderEmitter.addState = addState;
	stateProviderEmitter.go = (newStateName, parameters, options) => {
		options = extend(defaultOptions, options);
		const goFunction = options.replace ? router.replace : router.go;

		return promiseMe(makePath, newStateName, parameters, options)
			.then(goFunction, err => handleError(`stateChangeError`, err))
	};
	stateProviderEmitter.evaluateCurrentRoute = (defaultState, defaultParams) => promiseMe(makePath, defaultState, defaultParams).then(defaultPath => {
		router.evaluateCurrent(defaultPath);
	}).catch(err => handleError(`stateError`, err));
	stateProviderEmitter.makePath = (stateName, parameters, options) => pathPrefix + makePath(stateName, parameters, options);
	stateProviderEmitter.getActiveState = () => lastCompletelyLoadedState.get();
	stateProviderEmitter.stateIsActive = (stateName = null, parameters = null) => {
		const currentState = lastCompletelyLoadedState.get();
		const stateNameMatches = currentState.name === stateName
			|| currentState.name.indexOf(`${stateName }.`) === 0
			|| stateName === null;
		const parametersWereNotPassedIn = !parameters;

		return stateNameMatches
			&& (parametersWereNotPassedIn || Object.keys(parameters).every(key => `${ parameters[key] }` === currentState.parameters[key]))
	};

	const renderer = makeRenderer(stateProviderEmitter);

	destroyDom = denodeify(renderer.destroy);
	getDomChild = denodeify(renderer.getChildElement);
	renderDom = denodeify(renderer.render);

	return stateProviderEmitter
};

function getContentObject(stateResolveResultsObject, stateName) {
	const allPossibleResolvedStateNames = parse(stateName);

	return allPossibleResolvedStateNames
		.filter(stateName => stateResolveResultsObject[stateName])
		.reduce((obj, stateName) => extend(obj, stateResolveResultsObject[stateName]), {})
}

function redirector(newStateName, parameters) {
	return {
		redirectTo: {
			name: newStateName,
			params: parameters,
		},
	}
}

// { [stateName]: resolveResult }
function resolveStates(states, parameters) {
	const statesWithResolveFunctions = states.filter(isFunction(`resolve`));
	const stateNamesWithResolveFunctions = statesWithResolveFunctions.map(getProperty(`name`));

	const resolves = Promise.all(statesWithResolveFunctions.map(state => new Promise((resolve, reject) => {
		const resolveCb = (err, content) => err ? reject(err) : resolve(content);

		resolveCb.redirect = (newStateName, parameters) => {
			reject(redirector(newStateName, parameters));
		};

		const res = state.resolve(state.data, parameters, resolveCb);
		if (isThenable(res)) {
			resolve(res);
		}
	})));

	return resolves.then(resolveResults =>
		combine({
			stateName: stateNamesWithResolveFunctions,
			resolveResult: resolveResults,
		}).reduce((obj, result) => {
			obj[result.stateName] = result.resolveResult;
			return obj
		}, {}),
	)
}

module.exports = abstractStateRouter;
//# sourceMappingURL=bundle.js.map
